#include "header.h"
#include "classes.h"
#include "session.h"
#include "algoClass.h"

//class Session;
AlgoStatusClass::AlgoStatusClass( GraphClass *givenGraphInfo) {
    INDEX i;

    type = NONE;
    noNodes = givenGraphInfo->getNoNodes() ;
    Status = new int[noNodes]; // IN_TREE node means included in tree
    Father = new INDEX[noNodes]; // not used currently
    ExistStream = new STREAM[noNodes];
    JoinToNode = new INDEX[noNodes];
    JoinStream = new STREAM[noNodes];
    JoinHapCost = new HAP[noNodes];
    JoinCost = new COST[noNodes];
    graphInfo = givenGraphInfo;

    for (i=0; i<noNodes; i++) {
      Status[i] = NONE;
      Father[i] = NONE;   /// changed 18/5/96 from "i to NONE"
      ExistStream[i] = NONE;
      JoinToNode[i] = NONE;
      JoinHapCost[i] = Minus_INFINIT_double ;
      JoinStream[i] = NONE;
      JoinCost[i] = INFINIT_double;
    }
  };

AlgoStatusClass::~AlgoStatusClass() {
  // Do not delete the graph !!!!!
  // since it may be used in other sessino or algorithms
  delete []Status;
  delete []Father;
  delete []ExistStream;
  delete []JoinToNode;
  delete []JoinStream;
  delete []JoinHapCost;
  delete []JoinCost;
 
    }


INDEX AlgoStatusClass::getNextNode(){
    INDEX part; // Auxiliary
     // ---- Eligible node ----
    INDEX node; // eligible node
    HAP hap;
    COST cost;
    STREAM stream;
    // ---- Best node so far ------------
    INDEX nextNode;
    COST minCost;
    HAP maxHapCost;
    STREAM maxStream;
    COST maxCost;
    // ------ Aux ----
    Session *sessPtr;
    INDEX *partToNodeIdx;

    sessPtr = getCurrentSession();
    partToNodeIdx = sessPtr->getSubToWholeIndex();
    nextNode = NONE;
    maxHapCost = Minus_INFINIT_double;
    maxStream = NONE;
    maxCost = Minus_INFINIT_double;
    minCost = INFINIT_double;
    //// ------- ON-LINE PROBLEM ---------
    if (sessPtr->isOnline() == YES) {
      for(part=0; part< sessPtr->getNoPart(); part++) {
	// choose nodes in the sequence they were selected
	// in Session::makeSessPart() method
	node = sessPtr->getJoinSequeNode(part);
	// MICHALIS 29.1.98 node = sessPtr->partToNode(part);
	if (isReached(node) == NO)
	  return node;
      }
    } else
      {
	/// ------ OFF-LINE PROBLEM ---------
	for(part=0; part < sessPtr->getNoPart(); part++) {
	  // ---- Translate part idx to node idx----
	  node = sessPtr->partToNode(part);
	  if (isReached(node) == NO) {
	    hap = getJoinHapCost(node);
	    stream = getJoinStream(node);
	    cost = getJoinCost(node);
	    switch (getType()) {
	    case MAX_STREAM_NPF:
	      if ((stream > maxStream) ||
		  ((stream == maxStream) && (cost < maxCost))) {
		maxStream = stream;
		nextNode = node;
		maxCost = cost;
	      };
	      break;
	    case MIN_COST_NPF:
	      if (cost < minCost) {
		minCost = cost;
		nextNode = node;
	      };
	      break;
	    case OPT_STREAM_OPT_NPF:
	      if (hap > maxHapCost) {
		maxHapCost = getJoinHapCost(node);
		nextNode = node;
	      };
	      break;
	    default:
	      error("Algo:getNextNode() : Unspecified algorithm type");
	      break;
	    }

	  }
	}
      }
    return nextNode;
};

