// -*-C++-*-
// This file contains all the classes -- 25-4-97 Michalis
//#include "header.h"
//   This file gets included in the header.h
// Methods are in the classes.cc


class setElementSelection {
  int *nonChosen; // array of values non chosen yet
  int index;      // pointer of last non chose element

public:
  setElementSelection(int a, int b);
  ~setElementSelection();
 int  getNext();

};  // end of class

class Node { 
    NODEID nodeId;
};

class Edge {
    
  WEIGHT  weight;
  WEIGHT  undirWeight;
  NODEID  maxNode ;
  NODEID  minNode ;
  BOOL   mstEdge;
  BOOL   multiEdge;

public:

  Edge() {
    weight = NO_EDGE;
    undirWeight = NO_EDGE;
    maxNode = NONE  ;
    minNode = NONE ;
    mstEdge = NO;
    multiEdge = NO;
  };

  void setWeight(WEIGHT w) {
    weight = w;
  };

WEIGHT getUndirWeight() {
    return undirWeight;
  };

  void setUndirWeight(WEIGHT w){
    undirWeight = w;
  };

  BOOL isNO_EDGE(){
    if (weight == NO_EDGE)
      return YES;
    else
      return NO;
  };

  WEIGHT getWeight() {
    return weight;
  };

  BOOL compare(Edge other) {
    //---------- check weights ----------
    if (( weight == UNKNOWN) ||
	(other.weight == UNKNOWN) )
      return UNKNOWN;

    if (  weight > other.weight) return GREAT;
    if (  weight < other.weight) return LESS;

    //---------- if equal check min Node ----------
    if (  minNode > other.minNode) return GREAT;
    if (  minNode < other.minNode) return LESS;
    //---------- if equal check max Node ----------
    if (  maxNode > other.maxNode) return GREAT;
    if (  maxNode < other.maxNode) return LESS;
    return EQUAL;
  };

  // --- from WEIGHT_stru
  void copy(Edge w) {
    *this = w;
  };
  void show() {
    Out << " W: "<<weight<<" max: "<<maxNode;
    Out <<" min: "<<minNode<<NL;
  };
  void makeNO_EDGE() {
    weight = NO_EDGE;
    maxNode = NONE;
    minNode = NONE;
  };
  void makeUNKNOWN() {  // used in minMOE
    weight = UNKNOWN;
    maxNode = NONE;
    minNode = NONE;
  };
  void load(WEIGHT w, NODEID n1, NODEID n2){
    weight = w;
    maxNode = max(n1,n2);
    minNode = min(n1,n2);
  };

  BOOL isUNKNOWN(){
    if (weight == UNKNOWN)
      return YES;
    else return NO;
  };
};  /// class Edge

/* some predeclarations */


// ------------- Path Class ----------------------------------------//

class PathClass {
  NODEID * nodeArr; // array of the NODEID of the nodes of the path
  INDEX lastNode; // last added node in the path.  Length = lastNode+1
  INDEX counter; // aux var. to traverse the path
  WEIGHT distance; // Total distance of the path

  int status; // indicates whether path is trivial:  one node loop

public:
PathClass(){
   //nodeArr = new NODEID[len];
   status = NONE;
    lastNode = NONE; 
    distance = 0;
  };
PathClass(INDEX len) {
  status = NONE;
    nodeArr = new NODEID[len];
    lastNode = NONE; 
    distance = 0;
  };
void setDistance(WEIGHT c) { distance = c; };
WEIGHT getDistance() { return distance; };
void initNodeArr(INDEX len){ nodeArr = new NODEID[len];
  };

void setNodeArr(NODEID *arr){    nodeArr = arr;   };
BOOL isEmpty() {
    if (lastNode == NONE)
      return YES;
    else
      return NO;
  };

void makeEmpty(){   lastNode = NONE;  };
// void makeTrivial(){ status = TRIVIAL_PATH;};
//   BOOL isTrivial() {
//  if (status == TRIVIAL_PATH)
//    return YES;
//  else 
//    return NO;
//   };
void createNodeArr(INDEX len) { nodeArr = new NODEID[len];  };
void addNode(NODEID newnode) {
    if (lastNode != NONE)
      lastNode++;
    else 
      lastNode = 0;
    //debug2( newnode << " -- " ); 
    nodeArr[lastNode] = newnode;
  };
void print() {
  INDEX index;

  if (isEmpty() == YES) {
    Out << " PathClass::print():  Empty Path ";
  } else {
    Out << " PathClass::print(): " << NL;
    Out << "Path: " << nodeArr[0];
    for(index=1; index<=lastNode; index++) 
      Out <<  " -> " << nodeArr[index] ;
    Out << NL;
  };
};
void setCounter(INDEX value) {  counter = value; };
void incrCounter() {   counter++; };
void decrCounter(){   counter--; };
  // traversing the path is based on the counter //
NODEID  getNodeId(){   return (nodeArr[counter]); }; 
INDEX  getLastNodeIdx() {return lastNode; }; 
INDEX  getNoHops() {return lastNode; };
  //INDEX getActiveLength() { return lastNode+1 ; };
INDEX getNoNodes() { return lastNode+1 ; };
void initTraversal() {
  if (isEmpty() == YES) 
    Error << " initTraversal(): probable attempt to traverse empty path" << NL;
 setCounter(0); }

NODEID getNextNodeId() {
    NODEID node; 

    if (counter > lastNode) 
      return NONE;
    node = getNodeId();
    incrCounter();
    return node;
  };

void readInversePath(PathClass *path) {
  //   Initialize current path with given "path"
  //    but in reverse traversal

    INDEX i;
    INDEX nnodes, lastNode;

    //debugN2(" PathClass::readInversePath() ");
    nnodes = path->getNoNodes();
    lastNode = path->getLastNodeIdx();

    initNodeArr(nnodes);
    makeEmpty();
    path->setCounter(lastNode);
    for(i=0; i<nnodes; i++) {
      addNode(path->getNodeId());
      path->decrCounter();
    };
    //debugN2(" END ");
  };

};  // end of PathClass


// ----------------------------------------------------
// -------------- GraphClass -------------------------
// ----------------------------------------------------

class GraphClass{
public:  // public moved here for debugging
  Edge **graph; // graph
  double   maxAsymmetry; // maximum asymmetry 
  // The following matric will hold the paths i,j from node i to node j
  PathClass **ShortPathMat;
  INDEX  nodes;

  //public:


 // -------- in classes.cc -----------------
GraphClass(INDEX noNodes);
void allShortestPaths();
void  GraphClass::createEdge( INDEX n1, INDEX n2, WEIGHT w, double asym);
  // ----------------------- -----------------
  int makeAsymmetry( double asy);  // change weight of edges
  void  setMaxAsymmetry(double asy) { maxAsymmetry = asy ;};
  double getMaxAsymmetry(){ return maxAsymmetry;};

PathClass * getPathPtr(INDEX n1, INDEX n2) {
    return (&ShortPathMat[n1][n2]);
  };

void  setShortestPathWeight(INDEX n1, INDEX n2, WEIGHT weight) {
    getPathPtr(n1, n2)->setDistance(weight);
  };
WEIGHT getShortestPathWeight(INDEX n1, INDEX n2) {
    return (getPathPtr(n1, n2)->getDistance());
  };
Edge **getGraph() { return graph; };
Edge **getMatrix() { return graph; };
INDEX getNoNodes() { return nodes; };
Edge *getEdge(INDEX from, INDEX to) { return &graph[from][to]; };
}; // end Graph Class definition


class HapCostInfoClass {
public:
  COST cost;
  HAP  hap;
  HAP  hapCost;

  HapCostInfoClass() {
    cost = 0;
    hap = 0;
    hapCost = 0;
  }
print() {
        Out << "  Hap: " << hap <<"  Cost: " << cost;
	Out << "  HapCost: " << hapCost;
  };
};




