#include "header.h"
#include "classes.h"
#include "session.h"
#include "algoClass.h"

//extern BOOL adjacent(INDEX n1, INDEX n2);


//*********************************
// PROC: randIntInAB()
// DOES: returns an integer in [a,b]
// ADDED: 10/1/96 by petros
//*********************************
int randIntInAB(int a, int b)
{
    return (int) rint((drand48()*(b-a)+a)) ;
}

//*************************************
// PROC: randomPick()
// DOES: returns the next node to de added in the 
//       spanning tree in proc findST_random.
//       The selection is done randomly among the
//       unvisited nodes.
// ADDED: 10/1/96 by petros
//*************************************
INDEX randomPick(INDEX N, int *Status)
{
    int i, nUnv, indx ;
	INDEX *unv ;

	unv = new INDEX[N] ;
    
    // store unvisited nodes
    nUnv = 0 ;
    for( i = 0 ; i < N ; i++ )
		if( Status[i] != VISIT_MST ) 
			unv[nUnv++] = i ;
    if( nUnv == 0)
		return NONE ;
	else if ( nUnv == 1 )
		return unv[0] ;

    // get a random integer number in [0,nUnv-1]
    indx = randIntInAB(0,nUnv-1) ;
	
	delete []unv ;
    return unv[indx] ;
} // randomPick
    
//*********************************
// PROC: doNeigh()
// DOES:  Updates the Father, Distance, Label matrices
//        For the MST algorithm -- called by findMST
// ADDED: 
//*********************************
void doNeigh(Edge **Mat, INDEX node, INDEX N, int *Stat,
 WEIGHT *Dist, INDEX *Father, WEIGHT *Label) {
  // Function: Finds the new NEIGHboors and updates all NEIGH distances
  //     It changes Stat, Dist, Father
  int i;

     for (i=0; i<N; i++) {
       //debug("doNeigh: adj: " << adjacent_general(Mat, node, i));
       if ((adjacentMat(Mat, node, i) == YES) && (Stat[i] != VISIT_MST)) {
	  
	  Stat[i] = NEIGH_MST;
	  if ( compareWeight(Dist[i], Mat[node][i].getWeight()) == GREAT){
	    Father[i] = node;
	    Dist[i] =  Mat[node][i].getWeight();
	    Label[i] = Label[node] + Dist[i];
	  } // if compare
	  //debug("doNeigh: adjacents " << node << " and  " << i <<" Dist: "<<Dist[i]);
	} // if adjacent
      } // for
}


INDEX minDistNeigh(INDEX N, int *Stat, WEIGHT *Dist) {
  INDEX minNode = NONE;
  int i;
  WEIGHT minDist;

  minDist = INFINIT_int;
  for (i=0; i<N; i++) {
    if ((Stat[i] == NEIGH_MST) && (compareWeight(Dist[i], minDist) == LESS)){
      minNode = i;
      minDist = Dist[i];   // make sure that this will copy struct
      // if WEIGHT becomes struct.
    }
    //debug("minDistNeigh: " << i << "  Dist:  " << Dist[i]);
  }
  return minNode;
 }

//*********************************
// PROC: findMST()
// DOES:  Finds the MST of the graph "Mat" and
//        Returns the "Father" of each pointer
//              and the Label is the distnace source - node (i think)
// ADDED: michalis
//*********************************
int findMST(Edge **Mat, INDEX N, INDEX root, INDEX *Father, WEIGHT *Label){
  // function: Given Adj Mat, Nodes, root, return the MST tree  
  // (Father) and the Min Distances (Label) -- both are *new* from caller

  INDEX newnode;  // the new node that becomes VISITed
  int *Status;    // Node status: NONE, NEIGH, VISIT_MST
  WEIGHT *Dist; // distance from VISITED group 
  int i;
  BOOL STOP;
  int *junkPtr;
  // ------ Initialization of Matrices ------------

  Status = new int[N];
  Dist = new WEIGHT[N];  
  for (i=0; i<N; i++) {
    Status[i] = NONE;
    Dist[i] = INFINIT_int;
    Label[i] = INFINIT_int;
    Father[i] = i;
  }

  Dist[root] = 0;
  Label[root] = 0;
  doNeigh(Mat,root, N, Status, Dist, Father, Label);
  Status[root] = VISIT_MST;
  STOP = NO;
  while (STOP == NO){
    newnode = minDistNeigh(N, Status, Dist);
    //debug("findMST: newnode: " << newnode);
    if (newnode == NONE)
      STOP = YES;
    else {
      Status[newnode] = VISIT_MST;
      doNeigh(Mat, newnode, N, Status, Dist, Father, Label);
    } // else
  } // while

  // delete local matrices
  //  debug( "findMST: delete junkPtr");
  //delete junkPtr;

  //debug( "before delete Status: " << Status << " " << Status[0]<<" "<<Status[1]);
  delete []Status;
  delete []Dist;

 return EXEC_OK;

} // function findMST()

double oldCostBand(BAND u) {
 return  ((COST) exp( (double) (A_COST * u))) ; }

double costBand(BAND u) {
 return  ((COST)  ( u)) ; }

COST StreamBand(STREAM s) {
  //Out << " StreamBand(): pow(1+A_COST, s)="<< pow(1+A_COST, s) <<NL;
  return ((COST) A_COST * pow(STREAM_RATIO, s));
};

double costEdgeBand(Edge *edge, BAND u){
  // returns the cost  of having bandwidth u over edge of weight w
  WEIGHT w;

  w = edge->getWeight();
  return  ((COST)  w * costBand(u)) ;
}; 

double costEdgeStream(Edge *edge, STREAM s){
   return costEdgeBand(edge, StreamBand(s));
};

double hapCostPath(HAP hap, COST cost) {
	// gain per cost unit
 if (HAP_COST_FUNC == RELAT_GAIN_FUNC)
  return 1+(((A_HAPCOST_PATH * hap) - (B_HAPCOST_PATH  * cost) ) / (double) 			(cost) );
 else
  return ( (A_HAPCOST_PATH * hap) - (B_HAPCOST_PATH  * cost)  );
}

HAP hapCostStream(HAP hap, COST cost) {
  // this function is used in selecting the optimal stream
  // for a given destination
  // I left it the same as hapCostPath
return hapCostPath(hap, cost);
};

HAP hapCostTree(HAP hap, COST cost) {
  // The evaluation of the HapCOst for the tree
  if (HAP_COST_FUNC == RELAT_GAIN_FUNC){
   return 1+(( (A_HAPCOST_TREE * hap) - (B_HAPCOST_TREE * cost)) / (double) cost);
 }else 
   return (A_HAPCOST_TREE * hap) - (B_HAPCOST_TREE * cost);
   //return hapCostPath(hap, cost);
}


void doPathUpdateStream(INDEX N,STREAM  newStream, INDEX father,
		    INDEX *Path, STREAM* ExistStream) {
  INDEX i; 

  for(i=0;  i < N; i++) {
    ExistStream[Path[i]] = newStream;
    // debug(" doPathUpdateStream: i: " << i <<" node: "<< Path[i] );
  }
}

void initializeArr(INDEX *Arr, INDEX N) {
    INDEX i; 
    
  for(i=0;  i < N; i++) {
    Arr[i] = 0;
  }
};

double doUpdateStream(Session *sessionPtr, INDEX toAdd, STREAM sNew,
 BOOL  *Stat, STREAM *ExistStr, INDEX *Fath, INDEX fath) {
  INDEX i, v, fv; // auxiliary i is node, v is VISITED, fv is v's father
  double cost, minCost; // the added cost and the max over all possible joins
  STREAM  sV;  // stream that node v receives currently
  INDEX sessRoot;
  INDEX noPart  ;
 
  noPart = sessionPtr->getNoPart();
  sessRoot = sessionPtr->getSubSessRoot();
  //for(i=0;  i < (sessionPtr->getNoPart()); i++) {
 
  v = fath;
      
  while  ((ExistStr[v] < sNew) && ( v != sessRoot)) {
	ExistStr[v] = sNew;
	// debug(" doUpdateStream:    node: "<< v );
	v = Fath[v];
  }
};



COST NPFcostUpdatePath(AlgoStatusClass *Algo, INDEX nodeFrom, INDEX nodeTo,
		       STREAM stream, BOOL update) {
  //*********************************************
  //	PROC: NPFcostUpdatePath()
  //	DOES: Calculates the cost of joining b to a (a->b)
  //          where no connection was before
  //          IF update == YES 
  //            Updates the status of the added path nodes
  //            Updates: Status, Father, Exist Stream
  //          ATTENTION: We assume  non existing streams in the path
  //          Returns: the total cost of the path
  //  ADDED:  27/4/96
  //*********************************************
  INDEX nodeId, prevNodeId;
  COST totCost; // calculate total cost of added path
  Edge **graph;
  PathClass *path;
  GraphClass *graphInfo;
  Edge *edge;

  graphInfo = Algo->getGraphInfo();
  //  graph = Algo->getGraph();
  path = Algo->getPath(nodeFrom, nodeTo);
  totCost = 0;
  if (path->isEmpty() == YES) {
    Error << " NPFcostUpdatePath(): attempt to update empty path" << NL;
    return Minus_INFINIT_double;
  };
  //if (update == YES) {
    //debug4(" NPFcostUpdatePath():  Added Path: ");
    //path->print();
  //  };
  path->initTraversal();
  prevNodeId = path->getNextNodeId();  // first node is already in Tree
  nodeId = path->getNextNodeId();
  while (nodeId != NONE) {
    if ( update == YES) {
      Algo->putInTree(nodeId);  /// make IN_TREE unless VISIT_MST
      Algo->setFather(nodeId, prevNodeId);
      if ( stream < Algo->getExistStream(nodeId)) {
	error("NPFcostUpdatePath(): attempt to decrease the Exist Stream");
	error(" Exist: " << Algo->getExistStream(nodeId)<<" New: "<<stream); 
      }
      else {  // Increase (only)  stream
		Algo->setExistStream(nodeId, stream);
	  }
    };

    edge = graphInfo->getEdge(prevNodeId, nodeId);
    totCost += costEdgeStream(edge, stream);
    prevNodeId = nodeId;
    nodeId = path->getNextNodeId();
  };

  return totCost; // return value of total cost
};


COST NPFcostOfPath(AlgoStatusClass *Algo, INDEX nodeFrom, INDEX nodeTo,
		       STREAM stream) {
  //*********************************************
  //	PROC: NPFcostOfPath()
  //	DOES: Adding a new node with a path.
  //          Calculate the cost of the path a->b routed with "stream"
  //          a is in the multicast tree and b is not
  //          ATTENTION: We assume  no existing streams in the path
  //          RETURNS: the total cost of the path for "stream"
  //          COMMENT: very similar to NPFupdatePath()
  //  ADDED:  27/4/96
  //*********************************************
  BOOL noUpdate;
  COST cost;

  noUpdate = NO;
  cost = NPFcostUpdatePath(Algo, nodeFrom, nodeTo, stream, noUpdate);
  return cost;
};

COST NPFcostUpdateStreamIncrease(AlgoStatusClass *Algo,Session *sessionPtr,
				 INDEX node, STREAM stream, BOOL update) {
  //*********************************************
  //	PROC: NPFcostUpdateStreamIncrease()
  //	DOES: Calculate the cost of increasing the stream to "stream"
  //            of "node"
  //          This along the path  source -> node
  //          Existing streams of the path are considered
  //          UPDATE: IF update == YES the exisitng streams ExistStr 
  //          RETURNS: the total cost of the increase
  //  ADDED:  27/4/96
  //*********************************************
  INDEX i, v, fathV; // auxiliary i is node, v is VISITED, fv is v's father
  COST cost; // the added cost and the max over all possible joins
  STREAM  oldStream;  // stream that node v receives currently
  INDEX sessRoot;
  // read only variables pointing to AlgoStatusClass
  INDEX *Fath;
  STREAM *ExStr;
  //  Edge **graph;
  GraphClass *graphInfo;
  Edge *edge;
 
  ExStr = Algo->getExistStreamArr();
  Fath  = Algo->getFatherArr();
  //graph = Algo->getGraph();
  graphInfo = Algo->getGraphInfo();

  sessRoot = sessionPtr->getSessRoot(); // index of original graph
  // Find increase of stream cost in existing tree
  cost = 0;
  v = node;
  while  ((ExStr[v] < stream) && ( v != sessRoot)) {

    fathV = Fath[v];
    oldStream = ExStr[v];
    debug2("  L____>  v : " << v);
    debug2(" Fath v: " << fathV);
    debugN2( " node stream: " << stream << " v stream: " << ExStr[v]);
    // The following is correct, because i just add the higher stream
    edge = graphInfo->getEdge(fathV, v);
    cost += (costEdgeStream(edge, stream) -
	     costEdgeStream(edge, oldStream));

    //    cost += (costEdgeStream(edgeMat(graph, fathV, v), stream) -
    //	     costEdgeStream(edgeMat(graph, fathV, v), oldStream));
    // conditional update of stream
    if ( update == YES) 
      Algo->setExistStream(v, stream);

    v = fathV;
  } // while
  return cost;
};

COST  NPFcostOfStreamIncrease(AlgoStatusClass *Algo,Session *sessionPtr,
				 INDEX node, STREAM stream) {
  //*********************************************
  //	PROC: NPFcostOfStreamIncrease()
  //	DOES: Calculate the cost of increasing the stream that "node"
  //          currently receives to "stream"
  //          This along the path  source -> node
  //          Existing streams of the path are considered
  //          RETURNS: the total cost of the increase
  //  ADDED:  27/4/96
  //*********************************************
  BOOL noUpdate;
  COST cost; 

  noUpdate = NO;
 cost = NPFcostUpdateStreamIncrease(Algo, sessionPtr, node, stream, noUpdate);
  return cost;
};

void copyArr(INDEX *A, INDEX *B, INDEX N) {
  INDEX i;

  for(i=0; i<N; i++)
    B[i] = A[i];
}


//*********************************************
//	PROC: checkStreamUpgrade()
//	DOES: Check how profitable is to increase the stream
//          that "node" receives to "newStr"
//          Note: "node" is already in the tree
//          Existing streams of the path are considered
//          "thresh" is the threshold of the hapCostStream()   
//           that determines profitability
//          RETURNS: YES for profitable stream increase, NO otherwise
//  ADDED:  28/6/96
//*********************************************
BOOL checkNodeStreamUpgrade(AlgoStatusClass *Algo,Session *sessionPtr,
			INDEX node, STREAM newStream, HAP thresh) {
  COST addCost;
  HAP   addHap, addHapCost;
  STREAM oldStream;

  
  addCost = NPFcostOfStreamIncrease(Algo, sessionPtr, node, newStream) ;

  oldStream = Algo->getExistStream(node);
  addHap = sessionPtr->getNodeStreamHap(node, newStream) - 
    sessionPtr->getNodeStreamHap(node, oldStream);
  addHapCost = hapCostStream(addHap, addCost);
  if (addHapCost > thresh) {
    return YES;
  } else {
    return NO;
  }

} // end of checkStreamUpgrade() \\

//*********************************************
//	PROC: doStreamUpgrade()
//	DOES: Try to increase the streams using existing tree
//  ADDED:  27/6/96
//*********************************************
BOOL doStreamUpgrade(AlgoStatusClass *Algo,Session *sessionPtr) {
  BOOL upgrade;
  BOOL updateYES = YES;
  INDEX node, noNodes;
  STREAM newStream;
  HAP thresh = 0; // for the HapCost decision


  noNodes = sessionPtr->getNoNodes();
 
  do {
    upgrade = NO;
    for(node=0; node < noNodes; node++) {
      if (sessionPtr->isPartic(node) == YES) {
	newStream = Algo->getExistStream(node) + 1;
	// check if reqquired stream 
	if ((sessionPtr->isNodeReqStream(node, newStream) == YES)
	    &&
	    (checkNodeStreamUpgrade(Algo, sessionPtr,
				    node, newStream,  thresh)==YES)) {

	  debug3(" doStreamUpgrade: UPGRADING: n: "<<node);
	  debugN3(" newStream: "<<newStream);
          NPFcostUpdateStreamIncrease(Algo, sessionPtr,
				 node, newStream, updateYES);
	    // We assume that the algorithm has terminated
	    // so no ndeed to upadte the joinning Information
   //	  NPFstreamIncreaseJoinInfoUpdate(Algo, sessionPtr,  node);
	  upgrade = YES;
	}
      }
    }
  }  while(upgrade == YES);

}


// NAME: findOptStreams()
// DOES: finds the opt stream that maximizes the hapCostPath()
//  for all the non-visited nodes
//  the path is from node -> dest
// Alters Stream[i] for the unvisited nodes i
// I can also alter Father[i] 

COST NPFjoinCost(AlgoStatusClass *Algo, Session *sessionPtr,
				 INDEX node, INDEX dest, STREAM stream) {
  COST cost1; // cost of expanding (virgin) path
  COST  cost2; // cost of stream increase
  PathClass *path;
  

  path = Algo->getPath(node, dest);
  cost1 = NPFcostOfPath(Algo, node, dest, stream);
  cost2 = NPFcostOfStreamIncrease(Algo, sessionPtr, node,  stream);
  debug2(" NPFjoinCost():  " << node << " str: " << stream);
  debugN2(" dest: " << dest);
  debug2(" >> costOfPath:" <<cost1<< " costOfStreamIncrease " << cost2);
  debugN2(" cost: " << cost1+cost2);
  if (minusInfinitCost(cost1) ||
      minusInfinitCost(cost2))
    return Minus_INFINIT_double;
  else
    return (cost1+cost2);
};

// //****************** added by petros 11/1/96 ********************
HAP findMSTAlgo(Edge **Mat, INDEX N, Session *sessionPtr,
		INDEX *Father, WEIGHT *Label, BOOL algo){
  int i;
  Error << " findMSTAlgo(): dummy function" << NL;
  // dummy
}

HAP NPFoptStreamOfPath(AlgoStatusClass *Algo,Session *sessPtr,
			INDEX newNode, INDEX dest, STREAM *optStream) {
  //*********************************************
  //	PROC: 
  //	DOES: calculates opt stream for newNode -> dest
  //          Updates the  JoinInfo
  //          
  //          
  //    RETURNS: 
  //  ADDED:  27/4/96
  //*********************************************
  STREAM stream, bestStream;
  STREAM minStream, maxStream; // range of required streams for each node
  PathClass *path;
  COST cost, bestCost;
  HAP  hap;
  HAP hapCost, bestHapCost;
  
  bestStream = 0;
  //  bestHapCost = Minus_INFINIT_double;
  minStream = sessPtr->getNodeMinReqStream(dest);
  maxStream = sessPtr->getNodeMaxReqStream(dest);

  for(stream=minStream; stream<= maxStream; stream++) {

    cost = NPFjoinCost(Algo, sessPtr, newNode, dest, stream);
    if (minusInfinitCost(cost) == NO) {
    
      hap = sessPtr->getNodeStreamHap(dest, stream);
      hapCost = hapCostStream(hap, cost);

      if (hapCost > Algo->getJoinHapCost(dest) ) {
	//debug3(" NPFoptStreamOfPath(): UPDATE:");
	//debug3(" newNode:"<<newNode<<" dest: "<<dest);
	//debug3(" stream: "<<stream);
	//debug3(" OLD: stream: " << Algo->getJoinStream(dest));
	//debugN3(" toNode: " << Algo->getJoinToNode(dest));
	//debug3(" --->> Path: ");
	//      Algo->getPath( newNode, dest)->print();
	Algo->setJoinHapCost(dest, hapCost);
	Algo->setJoinToNode(dest, newNode);
	Algo->setJoinStream(dest, stream);
	Algo->setJoinCost(dest, cost);
	
      };
    };
  };
  
  (*optStream) = Algo->getJoinStream(dest);
  return Algo->getJoinHapCost(dest);
};

  // *********************************************
  //	PROC: NPFstatusUpdate()
  //	DOES: Update the status in Algo
  //          Calls NPFcostUpdatePath( ,yes) and
  //          NPFcostUpdateStreamIncrease( ,yes)
  // RETURNS: 
  //   ADDED:  1/5/96
  // *********************************************
void NPFstatusUpdate(AlgoStatusClass *Algo, Session *sessPtr, INDEX newNode){
  INDEX dest;
  INDEX noNodes;
  STREAM optStream;
  HAP hapCost;
  BOOL yesUpdate = YES;
  STREAM newStream;
  STREAM existStream; // stream node receives already (node maybe in Tree already
  INDEX nodeTo; // auxiliary node: nodeTo -> newNode is the possible path used
  
  // // debugN2(" NPFnewNodeUpdate(): start");
  noNodes = sessPtr->getNoNodes();
  // ----- Update Stream Info on Existing Tree nodes -------
  newStream = Algo->getJoinStream(newNode);
  existStream = Algo->getExistStream(newNode);
  if (existStream >= newStream) {
    // node is already in the tree with higher stream
    // do not allow consideration of less stream
   newStream = existStream;
  }
  if (newNode != sessPtr->getSessRoot()) {
    nodeTo = Algo->getJoinToNode(newNode);
    // ---- attention the path is nodeTo -> newNode
    NPFcostUpdatePath(Algo, nodeTo,  newNode,   newStream, yesUpdate);
    // hade it with newNode before -- changed 13/5/96 afternoon
    NPFcostUpdateStreamIncrease(Algo, sessPtr, nodeTo, newStream, yesUpdate);
  };
};

  // *********************************************
  //	PROC: NPFnodeJoinInfoUpdate()
  //	DOES: Update the join Info in "Algo"
  //          considering  only one "newNode"
  //          Calls NPFoptStreamOfPath() for every nonVISIT destination
  //   ADDED:  1/5/96
  // *********************************************
void NPFnodeJoinInfoUpdate(AlgoStatusClass *Algo, Session *sessPtr, INDEX newNode) {
  INDEX dest;
  INDEX noNodes;
  STREAM optStream;
  HAP hapCost;
  BOOL yesUpdate = YES;
  STREAM newStream;
  INDEX nodeTo; // auxiliary node: nodeTo -> newNode is the possible path used
  // // debugN2(" NPFnewNodeUpdate(): start");
  noNodes = sessPtr->getNoNodes();
  // ---------- Update Join Info of destinations---------
  for(dest=0; dest<noNodes; dest++) {
    if ((sessPtr->isPartic(dest) == YES)
	&& (Algo->isInTree(dest) == NO) ){
      //      	&& (Algo->is Reached(dest) == NO) ){   // changed 27/5/96
      // If dest is in Tree then only stream increase can take place
      NPFoptStreamOfPath(Algo, sessPtr, newNode, dest, &optStream);
    };
  };

};
// *********************************************
//	PROC: NPFstreamIncreaseJoinInfoUpdate()
//	DOES: Update the join Info in "Algo"
//            When hanging:  joinNode -> newNode
//          considering the  path of the  path root->joinNode
//   may have its stream increased 
//          For every  node calls NPFnodeJoinInfoUpdate()
//   ADDED:  13/5/96
// *********************************************
void NPFstreamIncreaseJoinInfoUpdate(AlgoStatusClass *Algo, Session *sessPtr, INDEX newNode) {

  INDEX node, root;

  // Let node be the node in the tree: node -> newNode
    root = sessPtr->getSessRoot();
  if (newNode == root) 
    return;
  else {
    node = Algo->getJoinToNode(newNode);
    
    while( node != root ) {
      debug3( "NFstreamIncreaseJoinInfoUpdate(): node of stream increased path "<< node<<" <-- ");
      NPFnodeJoinInfoUpdate(Algo, sessPtr, node);
      node = Algo->getFather(node);
    };
  };
    debugN2(" NPFstreamIncreaseJoinInfoUpdate(): end");
    // Most probably not needed
    // NPFnodeJoinInfoUpdate(Algo, sessPtr, root); 
}; // -- end of NPFsteramIncreaseJoinInfoUpdate()

  // *********************************************
  //	PROC: NPFpathJoinInfoUpdate()
  //	DOES: Update the join Info in "Algo"
  //          considering the new path that is added 
  //          For every path node calls NPFnodeJoinInfoUpdate()
  //   ADDED:  1/5/96
  // *********************************************
void NPFpathJoinInfoUpdate(AlgoStatusClass *Algo, Session *sessPtr, INDEX newNode) {
  INDEX fromNode; // beginning of the path , already in the Tree 
  PathClass *path;
  INDEX nodeId;
  
  
 
  if (newNode == sessPtr->getSessRoot()) {
    // -------  If the newNode is root no path exists ----------
    NPFnodeJoinInfoUpdate(Algo, sessPtr, newNode);
  } else {
    // --- Find Path ------
    fromNode = Algo->getJoinToNode(newNode);
    path = Algo->getPath(fromNode, newNode);
    path->initTraversal();
    //debug3(" NPFpathJoinInfoUpdate():  Added Path: ");
    //path->print();
    // ---  First node "fromNode" is already in the tree
    nodeId = path->getNextNodeId();
    if ((nodeId != fromNode) || (Algo->isInTree(newNode) == NO))
      Error<<"NPFpathJoinInfoUpdate: ERROR in JoinInfo update from path"<<NO;
    // nodeId = path->getNextNodeId();
    while (nodeId != NONE) {
      // --- Update JoinInfo for all the path nodes ------
      NPFnodeJoinInfoUpdate(Algo, sessPtr, nodeId);
      nodeId = path->getNextNodeId();
    };

    // I must have here the NPFnodeJoinInfoUpdate 
    // for the already tree nodes the since there may be a stream Increase
    // that may change things !!!!!!!!
  }; // -- else --  
};

HAP performNPF(AlgoStatusClass  *Algo, Session *sessionPtr,
		 BOOL algo, HapCostInfoClass *hapCostPtr) {
  // function: Given Adj Mat, Nodes, root, return the MST tree  
  // (Father) and the Min Distances (Label) -- both are *new* from caller
  //Edge **graph;   //graph
  //BOOL *Status;    // Node status: NONE, NEIGH, VISIT_MST
  //STREAM *ExistStream; // the stream a node currently has 
  //INDEX *Father;  // multicast tree father
  HapCostInfoClass *impHapCostPtr;

  
  COST totalCost;
  HAP totalHap;
  COST totalHapCost;
  int i;
  BOOL STOP;
  int *junkPtr, junk;
  INDEX root;
  INDEX noNodes;
  STREAM sMax; // max stream of session
  STREAM exiStream; // existing stream finally
  STREAM existStream; // stream node has when joining (maybe already in Tree)
  INDEX newNode;  // the new node that becomes VISITed
  INDEX  father; // the  father of the new node
  INDEX node; // auxnno
  INDEX noPart;
  INDEX currNoPart;  // counts current number of participants
  COST addedCost;  // cost of adding the newNode
  HAP  addedHap;   // happiness of adding the newNode
  HAP addedHapCost; // hapcost of adding the newNode
  STREAM newStream;  // the stream with which the newNode is added
  INDEX *Path;
  INDEX pathLen;
  NODEID toNode;    // node that current node will join to
  COST totStepHapCost;
  PathClass path;
  BOOL checkCor;
  COST cost; // auxiliary
  Edge *edge; // used to traverse the resulting tree
  HAP hap; // auxiliary for calculating the total hap
  INDEX *dumbTree;

  BOOL	alreadyInTree ; // aux: if newNode to join is in Tre already
  //                      because it belongs to a path of another node
  char  nametag[15]; 

  impHapCostPtr = new HapCostInfoClass;
  noPart = sessionPtr->getNoPart();
  noNodes = Algo->getNoNodes();
  debugN2(" NPF starting:  performNPF(): graph Ptr: " << Algo->getGraph() );
  // ------ Session Information Retrieved  ------------
  root = sessionPtr->getSessRoot();
  sMax = sessionPtr->getSessMaxStream();

  // --------- Initialization of variables ------------
  totalCost = 0;
  totalHap =0;
  totalHapCost =0;
  totStepHapCost = 0;
  currNoPart = 0;

  // ------ Initialization of Matrices ------------

  //  ---------- Intialize Algorithm Status ------------
  Algo->setSession(sessionPtr);
  newNode = root;
  Algo->setFather(root, root);
  Algo->setJoinToNode(root,root); // make father not NONE
  Algo->setJoinStream(root, sMax);
  Algo->setJoinHapCost(root, 0);
  Algo->setJoinCost(root, 0);
  alreadyInTree = NO;

  Out << " performNPF(): online flag is " ;
  Out << (Algo->getCurrentSession())->isOnline() << NL;
  if ( (Algo->getCurrentSession())->isOnline() == YES) 
    strcpy(nametag, "online_NPF");
  else
    strcpy(nametag, "NPF");

  STOP = NO;
  while (STOP == NO){
    if (newNode == NONE)
      STOP = YES;
    else {
	// -------- Update status of new node ---------
	Algo->setStatus(newNode, REACHED);
	currNoPart++ ;  // Increase number of participatns
	// --- find  the stream the node will get -----
	newStream = Algo->getJoinStream(newNode);
	existStream = Algo->getExistStream(newNode);
	//Make sure that stream increase only
	if (newStream <= existStream) {
	  // no need to update joining info
	  newStream = existStream;
	}
	//////// *************** ATTENTION ********	
	/// ******* stream must only increase ****************
	Algo->setExistStream(newNode, newStream);
      
	// ----- Keep track of hap and cost --------------------
	addedHapCost = Algo->getJoinHapCost(newNode);

	toNode = Algo->getJoinToNode(newNode);



	totalHapCost += addedHapCost;
	totalCost += Algo->getJoinCost(newNode);
	totalHap += sessionPtr->getNodeStreamHap(newNode, newStream);
	// ------- Update Status of participating nodes (path, tree part)
	NPFstatusUpdate(Algo, sessionPtr, newNode);
	// ------- Update Join Information of other destinations
	// ++++++ na balw kai streamIncrJoinInfo Update +++++
	// na checkarw 
	NPFstreamIncreaseJoinInfoUpdate(Algo, sessionPtr, newNode);
	if (alreadyInTree == NO) {
	  // ****************************************
	  // Na balw na krataw to max (ExistStr, newStr)?
	  // san JoinStream to new node?
	  NPFpathJoinInfoUpdate(Algo, sessionPtr, newNode);
	  pathLen = Algo->getPath(toNode, newNode)->getNoHops();
	 } else {
	   // node was already in tree
	   pathLen = 0;
	 }
	//if (newNode == root)
	//       Out << " Source-" ;
	//printHops(nametag, newNode, toNode, pathLen, currNoPart-1, noNodes);

	alreadyInTree = NO;
	//	newNode = NPFchooseNextNode(Algo);
	// CHANGE 26-4-97: Michalis
	newNode = Algo->getNextNode();
	//	Out << " performNPF(): newNode: " << newNode << NL;
	if (newNode != NONE && Algo->isInTree(newNode) == YES) {
	  alreadyInTree = YES;
	  /// temporarily  it has to be changed to stream upgrade function
	  Error << " Destination already in tree N: " << newNode ;
	  Error << " ExistStream: "<< Algo->getExistStream(newNode);
	  Error << " OptStream: " << Algo->getJoinStream(newNode)<< NL;
	  Algo->setStatus(newNode, REACHED);
	  
	};
	//	break;
	//}; // end of switch

    } // else
  } // while

  Out << " End of loop for algorithm " << NL;
      hapCostPtr->hap = totalHap;
      hapCostPtr->cost = totalCost;
      hapCostPtr->hapCost = hapCostTree(totalHap, totalCost);
      // Out << " totalCost= " << totalCost << NL;
      
      calcPathLengths( nametag, \
		       Algo->getGraphInfo() , \
		       sessionPtr, \
		       Algo->getFatherArr(), \
		       Algo->getExistStreamArr(), \
                       dumbTree );
      calculateTreeCost( Algo->getGraphInfo(), sessionPtr,
		Algo->getFatherArr(), Algo->getExistStreamArr(), hapCostPtr  );
      debugN3( " NPF: " );
      // check if all nodes are spanned by the tree
      checkCor = spansAllDestinations(sessionPtr,
				       Algo->getFatherArr(), root);
      if (checkCor == NO) 
	error(" NPF: Not all destinations were covered ");

      //  ---  STREAM UPGRADING -- DON'T DELETE
      //doStreamUpgrade( Algo, sessionPtr);
      // calculateTreeCost(Algo->getGraphInfo(), sessionPtr,
      //	Algo->getFatherArr(), Algo->getExistStreamArr(), impHapCostPtr);
      debugN3( " NPF after the stream Upgrade: " );
      debug3(" Improvement: Part: "<<noPart<<" Orig_HC: "<<hapCostPtr->hapCost);
      debugN3(" Imp_HC: "<< impHapCostPtr->hapCost);
      
  return hapCostPtr->hapCost;

} // function findMST()
