#include "header.h"
#include "classes.h"
#include "session.h"

  // ATTENTION : Here Dest in methods means the index of the complete graph

HAP Session::calcMaxTotalHap() {
    INDEX i;
    HAP totHap;
    
    totHap = 0;
    for(i=0; i<noPart; i++) {
	if (i != getSubSessRoot())
	totHap += getPartStreamHap(i, getDestMaxReqStream(i));
    }
    return totHap;
};

 
Edge** Session::makeCompleteGraph(Edge **graph, GraphClass *graphInfo){
  Edge **EdgeCompl;

    debug( "makeCompleteGraph() start");
    error(" NOT SUPPORTED: look in archive or unused.cc ");
    // completeSubgraph(graph,  nodes, this, graphInfo, EdgeCompl);
   debug( "makeCompleteGraph() end");
  return EdgeCompl;
 };
  // ----------- Constructor of Session ------------------
Session::Session(INDEX nonodes, INDEX part, STREAM s){
//        /*****************************************
// 	constructor of class Session
// 	nonodes = number of nodes,
// 	prob = probability - density of session, and
// 	s = max stream number
// 	*****************************************/

    int n,i;

    nodes = nonodes;
    Smax = s;
    setOnline(NO);   // By default session is off-line
    //thresh = prob;
    // In case
    if (part >  nodes) {
      noPart = nodes;
      error(" Participants cannot exceed  Nodes.");
    } else
      noPart = part;

    problemType = NON_FIXED;  // default value
    Partic = new BOOL[nodes];
    wholeToSubIndex = new INDEX[nodes];
    subToWholeIndex = new INDEX[nodes];
    MaxStreamArr = new STREAM[nodes];
    // debug("Constructor: Partic: " << Partic ); 

    NodeStreamHap = new (HAP *)[nodes] ;
    EdgeCompl = new (Edge *)[nodes];
    for(n=0; n<nodes; n++) {
      Partic[n] = NO;  // initially none participates
       NodeStreamHap[n] = new HAP[Smax+1];
       EdgeCompl[n] = new (Edge)[nodes];
      /* init matrix to zero */
      for(i=0; i<=Smax; i++)
	  NodeStreamHap[n][i] = 0;
    };
};
// ----------- Destructor of Session ------------------
Session::~Session() {
  INDEX n;

  delete []Partic;
  delete []wholeToSubIndex;
  delete []subToWholeIndex;
  delete []MaxStreamArr;
  for(n=0; n<nodes; n++) {
    delete []NodeStreamHap[n];
    delete []EdgeCompl[n];
  }
  delete []NodeStreamHap;
  delete []EdgeCompl;
};  /// -- end of ~Session

HAP Session::hapVal(STREAM s) {
  // assuming that happiness value depends only on the stream
  if (s < 0) 
    cerr << "hapVal(): stream was less than zero: " << s;
  // OLD version return ( (HAP)  A_HAPVAL * s +  C_HAPVAL);
  return ( (HAP)  B_HAPVAL * log((double)  A_HAPVAL * s + 1) + C_HAPVAL );
};

void Session::print(){
  INDEX part;
  STREAM str;

  for(part = 0; part < getNoPart(); part++) {
    Out << " part: "<< part << " NodeIdx: " << partToNode(part);
    Out << " MaxStream: " << getDestMaxReqStream(part);
    // for(str=0; str < getDestMaxReqStream(part); str++)
    //Out<<" str: "<< str<<" hap: "<<getPartStreamHap(part, str);
    Out << NL;
  }
}

void Session::makeStreamHap() {

  // *****************************************
  //     method of Session 
  //    assigns  streams ONLY to participants
  //   and happinnes values
  //   NodeStreamHap matrix has entrie only for participants !!!
  //   *****************************************
     int ni;
     STREAM si;
     STREAM s_max; //        maximum stream for node i =/=  Smax 

     // --- initilize random generator with a constant
     srand48(SESSION_SEED) ;
     maxReqStream = 0; /* intialize max required stream */
     for(ni=0; ni<noPart; ni++) {
	 s_max = (STREAM) chooseInt(0, Smax);
	 //  keep track of maximum required stream of all nodes 
	 if (s_max >= maxReqStream) 
	     maxReqStream = s_max;
	 MaxStreamArr[ni] = s_max;

	 debug3(" part: "<< ni << " NodeIdx: " << partToNode(ni) );
	  debugN3( " Smax: " << s_max);
	 for(si=0; si<=Smax; si++) {
// 	      define happiness for ni si 
// 	        for the moment  all partici have equal happiness
// 	        for their required satreams 
// 	      NON FIXED : define happiness for ni si 
// 	        for the moment  all partici have equal happiness
// 	        for their required satreams 
// 	      the stream increase until s_max 
	     if (si == 0) {
		 NodeStreamHap[ni][si] = hapVal(si);
	     } else {
		 if (si <= s_max) {
		     NodeStreamHap[ni][si] =
		       NodeStreamHap[ni][si-1] + hapVal(1);
		 } else
		     NodeStreamHap[ni][si] = NodeStreamHap[ni][s_max]; 
	     }
	      debug3(" s: "<< si <<" h: " << NodeStreamHap[ni][si] <<" " );
	 };// for(si=0; si<=s_max; si++) 
	  debugN3(" ");
	  // };   if (isPartic(ni) == YES) 
     };
 };

BOOL Session::validPartIdx(INDEX idx){
  // Checks if Index idx is in [0, noPart-1]
    if (idx < 0){
      error("Session::validPartIdx(): Negative Index");
      return NO;
    }
    else if ( idx > noPart-1) {
      error("Session::validPartIdx(): Index is too High");
      return NO;
    } else
      return YES;
      
  };
NODEID Session::getJoinSequeNode(INDEX part) {
  // Return the part-ith node in a online problem
  // the nodes are selected

  
  if (validPartIdx(part) == NO)
    {
      error("AlgoStatusClass::getJoinSequeNode:");
      return NONE;
    }
  return partToNode(part);
  
};

void Session::makeSessPart() {
  //  *****************************************
  // method of Session
  // selects the participants and the sess root
  //       alters  Session::Partic[] 
  //  creates Session:: wholeToSubIndex and subToWholeIndex
  /// *****************************************
 
  INDEX  i, maxNode = NONE;  // maxNode is node with Max prob
  PROB  prob, maxProb; // variables to help select the root
  INDEX partCount; // counting participants
  INDEX totPart, totNodes; // total number of participants and nodes
  INDEX newPart;  // the new participant
  setElementSelection *partSelect;

  // debug("makeSessPart: PArtic: " << Partic );
  srand48(SESSION_SEED) ;   // re-initalize random generator
  partCount = 0;
  maxProb = -1; // negetive
  totPart = getNoPart();
  totNodes = getNoNodes();
  partSelect = new setElementSelection(0, totNodes-1);
  while (partCount < totPart) {
    // the choice can be made faster
    // check the bounds should it be totNodes-1
    //    newPart = chooseInt(0, totNodes);
    newPart = partSelect->getNext();
    
    if (isPartic(newPart) == NO) {
      Partic[newPart] = YES;
      // wholeToSubIndex[] is the order of selection
      //      for destinations
      //      It can be as  the Online Order
      wholeToSubIndex[newPart] = partCount;
      subToWholeIndex[partCount] = newPart;
      partCount++;
      //  select the root with max prob 
      prob = (float) drand48() ; // number in  [0 , 1]
      // select root to be the first chosen node
      if ( partCount == 1) {
	sessRoot = newPart ;
      };
    };

  };
} //  end of method Session::makeSessPart()

